# python
# reannotate.py
# Author: Warren Kaplan, Kevin Ying, Mark Cowley
# Date:   13/6/2014

# Reannotate a GCT file with a CHIP format file

import sys
import os
from optparse import OptionParser
#import urllib
from urlparse import urlparse
import ftplib
import socket

# default debug variable
debug = False

MODULE_VERSION = "7.7"
MODULE_NAME = "ReannotateGCT"

def main():
    global debug

    usage =  "Usage: %s -g <gct file> -c <chip file> -o <output file> [-v]" % os.path.basename(sys.argv[0])
    parser = OptionParser(usage=usage)
    parser.add_option("-g", "--gctFile",
              action="store", type="string", dest="gctFile",
              help="The GCT file to be modified")
    parser.add_option("-c", "--chipFile",
              action="store", type="string", dest="chipFile",
              help="The CHIP file to be modified")
    # A custom CHIP file. Overides chipFile Used by the GenePattern module. can ignore otherwise
    parser.add_option("-d", "--custChipFile",
              action="store", type="string", dest="custChipFile",
              help="A custom CHIP file, overrides --chipFile if specified #Used for GenePattern module#")
    parser.add_option("-o", "--outFile",
              action="store", type="string", dest="outFile",
              help="The name of the annotated GCT file to write to")
    parser.add_option("-v", "--verbose",
              action="store_true", dest="verbose",
              help="Make module talkative")
    (options, args) = parser.parse_args()

    chipFile = options.chipFile
    if options.custChipFile != None:
        chipFile = options.custChipFile
    gctFile = options.gctFile
    outFile = options.outFile
    verbose = options.verbose

    if verbose:
        sys.stdout.write("You are running the " + MODULE_NAME + " module (version " + MODULE_VERSION + ") on " +
                         socket.gethostname() + "\n")
        sys.stdout.write("Reannotating GCT file: " + gctFile + "\n")
        sys.stdout.write("Using descriptions from CHIP file: " + chipFile + "\n")
        sys.stdout.write("Saving results to GCT file: " + outFile + "\n")

    if outFile == os.path.basename(gctFile):
        sys.stderr.write("You must choose a different name for your output GCT file\n")
        exit(1)

    # Open output file handle
    outf = open(outFile, 'w')

    # Store both files
    (header, origGCT)  = getOrigGCTList(gctFile)
    chipDict = getChipDict(chipFile)

    # Print header
    for i in header:
        print >>outf, "\t".join(i)

    # Loop through gct and annotate as necessary
    notFound = []
    for i in range(len(origGCT)):
        #print originalGCTlist[i]
        probe = origGCT[i][0]
        if chipDict.has_key(probe):
            # Update the Description field
            origGCT[i][1] = chipDict[probe]
            print >>outf, "\t".join(origGCT[i])
        if probe not in chipDict:
            notFound.append(probe)
            #origGCT[i][1] = "Probe Annotation Not Updated with latest Chip File"
            print >>outf, "\t".join(origGCT[i])

    outf.close()

    if len(notFound) > 0:
        print (str(len(notFound)) + " out of " + str(len(origGCT)) + " probes were not found in the chip file\nThe probes were:")
        print (notFound)
    

def getChipDict(file):
    chipDict = {}
    # get the file handle
    f = getReadHandle(file)
    #f = open(file, 'rU') # "rU" allows for reading from unix, windows and mac formats
    # parse chip file headers
    header = f.next().strip().split("\t")
    if (debug):
        print header
    assert (len(header) >= 3)
    #sys.stderr.write(str(header) + "\n")

    for line in f:
        line = line.strip()
        line = line.split("\t")
        #print line
        if len(line) >= 3: # Rudimentary checking of line format
            probe = line[0]
            genesymbol = line[1]
            description = line[2]
            chipDict[probe] = genesymbol + " :: " + description
        elif len(line) == 2:
            probe = line[0]
            genesymbol = line[1]
            chipDict[probe] = genesymbol
        else:
            sys.stderr.write("ChipFile: line ignored, bad format:" + str(line) + "\n")
#AFwefawefegrhsetrjydtdrh

    f.close()
    #print chipDict
    return chipDict

def getOrigGCTList(file):
    originalGCTlist = []

    
    # Open and save gct file
    # check for existence of file on local disk
    if (not os.path.isfile(file)):
        sys.stderr.write("The module can't access the GCT file")
    # Get the file handle
    f = getReadHandle(file)
    # f = open(file, 'rU') # "rU" allows for reading from unix, windows and mac formats

    # parse GCT file data and headers
    version = f.next().strip().split("\t")
    nrowSample = f.next().strip().split("\t")
    colnames = f.next().strip().split("\t")
    headers = [version, nrowSample, colnames]

    for line in f:
        line = line.strip()
        line = line.split("\t")
        if len(line) == (len(colnames)): # Rudimentary checking of line format
            #print line
            originalGCTlist.append(line)
        else:
            sys.stderr.write("GCTFile: line ignore, bad format:" + str(line) + "\n")
    f.close()
    if len(originalGCTlist) != int(nrowSample[0]):
        sys.stderr.write("Warning: " + str(nrowSample[0]) + " does not match actual no. of rows " + str(len(originalGCTlist)) + '\n')
    return (headers, originalGCTlist)

# Gets the file handle from a path
# Should handle url's such as ftp://*
#   as well as /path/to/file/ paths
def getReadHandle (url):
    # Parse the url
    if (debug):
        sys.stderr.write("grabbing read handle for url: " + url + "\n")
    urlp = urlparse(url)
    if (debug):
        sys.stderr.write(urlp.path + "\t" + urlp.scheme + "\n")

    if urlp.scheme == 'ftp':
        # note if run via GenePattern server, this code is never executed as GP auto-downloads FTP URLs and puts data
        #  in the jobResults/<jobID> folder.
        if (debug):
            sys.stdout.write("Attempting to grab ftp file: " + url)
        (wd, filename) = os.path.split(urlp.path)

        # read ftp file
        ftp = ftplib.FTP()
        ftp.connect(urlp.hostname,urlp.port)
        ftp.login() # login as anonymous
        ftp.cwd(wd)
        tmpLocalFileName = filename
        tmpLocalFile = open(tmpLocalFileName, 'wb')
        ftp.retrlines('RETR ' + tmpLocalFileName, lambda d: tmpLocalFile.write(d + "\n"))
        tmpLocalFile.close()

        return open(tmpLocalFileName, 'rU')
    elif urlp.scheme == 'file' or urlp.scheme == '':
        if (debug):
            sys.stderr.write("Attempting to open file: " + urlp.path +'\n')
        if os.path.exists(urlp.path) and os.path.isfile(urlp.path) and os.access(urlp.path, os.R_OK) and \
                os.path.getsize(urlp.path) > 0:
            if (debug):
                sys.stdout.write("File has size: " + str(os.path.getsize(urlp.path)) + "\n")
            return open(urlp.path, 'rU')
        else:
            sys.stderr.write("File does not exist, or is not readable: " + urlp.path + "\n")
            exit(1)
    else:
        if (debug):
            sys.stderr.write("Attempting to open URL: " + urlp.path +'\n')
        return open(urlp.path, 'rU')

if __name__ == "__main__":
    main()
